@@ -1,7 +1,7 @@
 import asyncio
 import base64
-import binascii
 import collections
+import functools
 import hashlib
 import itertools
 import logging
@@ -17,6 +17,8 @@ import msgpack
 from channels.exceptions import ChannelFull
 from channels.layers import BaseChannelLayer
 
+from .utils import _consistent_hash
+
 logger = logging.getLogger(__name__)
 
 AIOREDIS_VERSION = tuple(map(int, aioredis.__version__.split(".")))
@@ -42,14 +44,15 @@ def _wrap_close(loop, pool):
 class ConnectionPool:
     """
     Connection pool manager for the channel layer.
-
     It manages a set of connections for the given host specification and
     taking into account asyncio event loops.
     """
 
     def __init__(self, host):
-        self.host = host
+        self.host = host.copy()
+        self.master_name = self.host.pop("master_name", None)
         self.conn_map = {}
+        self.sentinel_map = {}
         self.in_use = {}
 
     def _ensure_loop(self, loop):
@@ -67,16 +70,27 @@ class ConnectionPool:
 
         return self.conn_map[loop], loop
 
+    async def create_conn(self, loop):
+        # One connection per pool since we are emulating a single connection
+        kwargs = {"minsize": 1, "maxsize": 1, **self.host}
+        if not (sys.version_info >= (3, 8, 0) and AIOREDIS_VERSION >= (1, 3, 1)):
+            kwargs["loop"] = loop
+        if self.master_name is None:
+            return await aioredis.create_redis_pool(**kwargs)
+        else:
+            kwargs = {"timeout": 2, **kwargs}  # aioredis default is way too low
+            sentinel = await aioredis.sentinel.create_sentinel(**kwargs)
+            conn = sentinel.master_for(self.master_name)
+            self.sentinel_map[conn] = sentinel
+            return conn
+
     async def pop(self, loop=None):
         """
         Get a connection for the given identifier and loop.
         """
         conns, loop = self._ensure_loop(loop)
         if not conns:
-            if sys.version_info >= (3, 8, 0) and AIOREDIS_VERSION >= (1, 3, 1):
-                conn = await aioredis.create_redis(**self.host)
-            else:
-                conn = await aioredis.create_redis(**self.host, loop=loop)
+            conn = await self.create_conn(loop)
             conns.append(conn)
         conn = conns.pop()
         if conn.closed:
@@ -95,11 +109,11 @@ class ConnectionPool:
             conns, _ = self._ensure_loop(loop)
             conns.append(conn)
 
-    def conn_error(self, conn):
+    async def conn_error(self, conn):
         """
         Handle a connection that produced an error.
         """
-        conn.close()
+        await self._close_conn(conn)
         del self.in_use[conn]
 
     def reset(self):
@@ -107,20 +121,31 @@ class ConnectionPool:
         Clear all connections from the pool.
         """
         self.conn_map = {}
+        self.sentinel_map = {}
         self.in_use = {}
 
+    async def _close_conn(self, conn, sentinel_map=None):
+        if sentinel_map is None:
+            sentinel_map = self.sentinel_map
+        if conn in sentinel_map:
+            sentinel_map[conn].close()
+            await sentinel_map[conn].wait_closed()
+            del sentinel_map[conn]
+        conn.close()
+        await conn.wait_closed()
+
     async def close_loop(self, loop):
         """
         Close all connections owned by the pool on the given loop.
         """
         if loop in self.conn_map:
             for conn in self.conn_map[loop]:
-                conn.close()
-                await conn.wait_closed()
+                await self._close_conn(conn)
             del self.conn_map[loop]
 
         for k, v in self.in_use.items():
             if v is loop:
+                await self._close_conn(k)
                 self.in_use[k] = None
 
     async def close(self):
@@ -128,15 +153,14 @@ class ConnectionPool:
         Close all connections owned by the pool.
         """
         conn_map = self.conn_map
+        sentinel_map = self.sentinel_map
         in_use = self.in_use
         self.reset()
         for conns in conn_map.values():
             for conn in conns:
-                conn.close()
-                await conn.wait_closed()
+                await self._close_conn(conn, sentinel_map)
         for conn in in_use:
-            conn.close()
-            await conn.wait_closed()
+            await self._close_conn(conn, sentinel_map)
 
 
 class ChannelLock:
@@ -179,6 +203,19 @@ class UnsupportedRedis(Exception):
     pass
 
 
+class BoundedQueue(asyncio.Queue):
+    def put_nowait(self, item):
+        if self.full():
+            # see: https://github.com/django/channels_redis/issues/212
+            # if we actually get into this code block, it likely means that
+            # this specific consumer has stopped reading
+            # if we get into this code block, it's better to drop messages
+            # that exceed the channel layer capacity than to continue to
+            # malloc() forever
+            self.get_nowait()
+        return super(BoundedQueue, self).put_nowait(item)
+
+
 class RedisChannelLayer(BaseChannelLayer):
     """
     Redis channel layer.
@@ -226,7 +263,9 @@ class RedisChannelLayer(BaseChannelLayer):
         # Event loop they are trying to receive on
         self.receive_event_loop = None
         # Buffered messages by process-local channel name
-        self.receive_buffer = collections.defaultdict(asyncio.Queue)
+        self.receive_buffer = collections.defaultdict(
+            functools.partial(BoundedQueue, self.capacity)
+        )
         # Detached channel cleanup tasks
         self.receive_cleaners = []
         # Per-channel cleanup locks to prevent a receive starting and moving
@@ -246,6 +285,7 @@ class RedisChannelLayer(BaseChannelLayer):
             raise ValueError(
                 "You must pass a list of Redis hosts, even if there is only one."
             )
+
         # Decode each hosts entry into a kwargs dict
         result = []
         for entry in hosts:
@@ -544,7 +584,11 @@ class RedisChannelLayer(BaseChannelLayer):
         Returns a new channel name that can be used by something in our
         process as a specific channel.
         """
-        return "%s.%s!%s" % (prefix, self.client_prefix, uuid.uuid4().hex,)
+        return "%s.%s!%s" % (
+            prefix,
+            self.client_prefix,
+            uuid.uuid4().hex,
+        )
 
     ### Flush extension ###
 
@@ -642,10 +686,12 @@ class RedisChannelLayer(BaseChannelLayer):
 
         for connection_index, channel_redis_keys in connection_to_channel_keys.items():
             # Discard old messages based on expiry
+            pipe = connection.pipeline()
             for key in channel_redis_keys:
-                await connection.zremrangebyscore(
+                pipe.zremrangebyscore(
                     key, min=0, max=int(time.time()) - int(self.expiry)
                 )
+            await pipe.execute()
 
             # Create a LUA script specific for this connection.
             # Make sure to use the message specific to this channel, it is
@@ -756,7 +802,7 @@ class RedisChannelLayer(BaseChannelLayer):
             # Get its redis key
             channel_key = self.prefix + channel_non_local_name
             # Have we come across the same redis key?
-            if channel_key not in channel_key_to_message.keys():
+            if channel_key not in channel_key_to_message:
                 # If not, fill the corresponding dicts
                 message = dict(message.items())
                 message["__asgi_channel__"] = [channel]
@@ -769,9 +815,9 @@ class RedisChannelLayer(BaseChannelLayer):
                 channel_key_to_message[channel_key]["__asgi_channel__"].append(channel)
 
         # Now that we know what message needs to be send on a redis key we serialize it
-        for key in channel_key_to_message.keys():
+        for key, value in channel_key_to_message.items():
             # Serialize the message stored for each redis key
-            channel_key_to_message[key] = self.serialize(channel_key_to_message[key])
+            channel_key_to_message[key] = self.serialize(value)
 
         return (
             connection_to_channel_keys,
@@ -813,15 +859,7 @@ class RedisChannelLayer(BaseChannelLayer):
     ### Internal functions ###
 
     def consistent_hash(self, value):
-        """
-        Maps the value to a node value between 0 and 4095
-        using CRC, then down to one of the ring nodes.
-        """
-        if isinstance(value, str):
-            value = value.encode("utf8")
-        bigval = binascii.crc32(value) & 0xFFF
-        ring_divisor = 4096 / float(self.ring_size)
-        return int(bigval / ring_divisor)
+        return _consistent_hash(value, self.ring_size)
 
     def make_fernet(self, key):
         """
@@ -866,7 +904,7 @@ class RedisChannelLayer(BaseChannelLayer):
 
         async def __aexit__(self, exc_type, exc, tb):
             if exc:
-                self.pool.conn_error(self.conn)
+                await self.pool.conn_error(self.conn)
             else:
                 self.pool.push(self.conn)
             self.conn = None
